package geniusweb.protocol.session.amop;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.EndNegotiation;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.actions.Vote;
import geniusweb.actions.Votes;
import geniusweb.inform.Agreements;
import geniusweb.issuevalue.Bid;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.progress.ProgressTime;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConn;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnections;
import geniusweb.protocol.session.amop.AMOPState.Phase;
import geniusweb.references.PartyWithProfile;
import tudelft.utilities.junit.GeneralTests;

public class AMOPStateTest extends GeneralTests<AMOPState> {
	private final long NOW = 1000;
	Bid a = mock(Bid.class), b = mock(Bid.class), c = mock(Bid.class),
			d = mock(Bid.class);

	private final PartyId party1 = new PartyId("party1");
	private final PartyId party2 = new PartyId("party2");
	private final PartyId party3 = new PartyId("party3");
	private final PartyId party4 = new PartyId("party4");
	private final List<Action> actions = new LinkedList<>();
	private final Action action = mock(Action.class);
	private final List<Action> actions2 = Arrays.asList(action);

	private final ProtocolToPartyConn party1conn = mock(
			ProtocolToPartyConn.class);
	private final ProtocolToPartyConn party2conn = mock(
			ProtocolToPartyConn.class);
	private final ProtocolToPartyConn party3conn = mock(
			ProtocolToPartyConn.class);

	private final ProtocolToPartyConnections connections = new ProtocolToPartyConnections(
			Arrays.asList(party1conn, party2conn));
	private final ProtocolToPartyConnections connections2 = new ProtocolToPartyConnections(
			Arrays.asList(party1conn, party3conn));
	private final ProtocolToPartyConnections connections3 = new ProtocolToPartyConnections(
			Arrays.asList(party1conn, party2conn, party3conn));

	private final Progress progresstime = mock(ProgressTime.class);
	private final ProgressRounds progressrounds = mock(ProgressRounds.class);
	private final ProgressRounds progressrounds1 = mock(ProgressRounds.class);
	private final AMOPSettings settings = mock(AMOPSettings.class);
	private final AMOPSettings settings2 = mock(AMOPSettings.class);
	private AMOPState state1, state1a, state2, state3, state4, state5, state6,
			state7, state8, state9;

	private final Bid bid1 = mock(Bid.class);
	private final Bid otherbid = mock(Bid.class);
	private final Accept accept1 = new Accept(party1, bid1);
	private final Accept accept2 = new Accept(party2, bid1);
	private final Accept acceptother = new Accept(party2, otherbid);
	private final Offer offer1 = new Offer(party1, bid1);
	private final EndNegotiation endNegotiation1 = new EndNegotiation(party1);

	private final Map<PartyId, PartyWithProfile> partyprofiles1 = mock(
			HashMap.class);
	private final Map<PartyId, PartyWithProfile> partyprofiles2 = mock(
			HashMap.class);

	private final Agreements agreements1 = new Agreements();
	private final Agreements agreements2 = mock(Agreements.class);

	private final Map<PartyId, ProtocolException> exceptions1 = mock(
			HashMap.class);
	private final Map<PartyId, ProtocolException> exceptions2 = mock(
			HashMap.class);

	private final List<PartyId> walkedaway0 = Arrays.asList();
	private final List<PartyId> walkedaway1 = Arrays.asList(party1);
	private final List<PartyId> walkedaway3 = Arrays.asList(party3);

	@Before
	public void before() {
		when(party1conn.getParty()).thenReturn(party1);
		when(party2conn.getParty()).thenReturn(party2);
		when(party3conn.getParty()).thenReturn(party3);
		when(party1conn.toString()).thenReturn("conn1");
		when(party2conn.toString()).thenReturn("conn2");
		when(party3conn.toString()).thenReturn("conn3");
		when(action.toString()).thenReturn("act1");

		when(progressrounds.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds1.getTerminationTime())
				.thenReturn(new Date(System.currentTimeMillis() + 3600000l));
		when(progressrounds.advance()).thenReturn(progressrounds1);

		state1 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state1a = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state2 = new AMOPState(Phase.INIT, actions2, connections, progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state3 = new AMOPState(Phase.INIT, actions, connections2, progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state4 = new AMOPState(Phase.INIT, actions, connections, progressrounds,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state5 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings2, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state6 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles2, agreements1, exceptions1,
				walkedaway0);
		state7 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles1, agreements2, exceptions1,
				walkedaway0);
		state8 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles1, agreements1, exceptions2,
				walkedaway0);
		state9 = new AMOPState(Phase.INIT, actions, connections, progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway3);
	}

	@Override
	public List<List<AMOPState>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(state1, state1a),
				Arrays.asList(state2), Arrays.asList(state3),
				Arrays.asList(state4), Arrays.asList(state5),
				Arrays.asList(state6), Arrays.asList(state7),
				Arrays.asList(state8), Arrays.asList(state9));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		// being lazy, toString is terrible large and everything's mocked
		return Arrays.asList("AMOPState.*", "AMOPState.*", "AMOPState.*",
				"AMOPState.*", "AMOPState.*", "AMOPState.*", "AMOPState.*",
				"AMOPState.*", "AMOPState.*");
	}

	@Test
	public void constructWith0Connection() {
		state1 = new AMOPState(Phase.INIT, actions,
				new ProtocolToPartyConnections(Arrays.asList()), progresstime,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
	}

	@SuppressWarnings("unused")
	@Test
	public void constructWithNullConnection() {
		// is allowed
		new AMOPState(Phase.INIT, actions, null, progresstime, settings,
				partyprofiles1, agreements1, exceptions1, walkedaway0);
	}

	@Test
	public void isInitFinalTest() {
		assertFalse(state1.isFinal(NOW));
	}

	@Test
	public void isInitFinalTest2() {
		assertFalse(state1.isFinal(NOW + 12000));
	}

	@Test
	public void withExceptionTest() {
		AMOPState newstate1 = state1.with(party1,
				mock(ProtocolException.class));
		// party errors are not terminating in AMOP
		assertFalse(newstate1.isFinal(NOW));
		AMOPState newstate2 = newstate1.with(party2,
				mock(ProtocolException.class));
		assertFalse(newstate2.isFinal(NOW));
	}

	@Test
	public void isOfferAgreementTest1() {
		List<Action> actions = Arrays.asList(offer1);
		AMOPState state = new AMOPState(Phase.INIT, actions, connections3,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		assertTrue(state.getAgreements().getMap().isEmpty());
	}

	@Test(expected = ProtocolException.class)
	public void illegalActionInIllegalState() throws ProtocolException {
		// party can not do accept. And certainly not in INIT phase.
		state1.with(party1, offer1).with(party1, accept2);
	}

	@Test
	public void isOfferAllowedInOFFER() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, offer1);
	}

	@Test(expected = ProtocolException.class)
	public void isOfferAllowedInINIT() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.INIT, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, offer1);
	}

	@Test(expected = ProtocolException.class)
	public void isOfferAllowedInVOTE() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.VOTE, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, offer1);
	}

	@Test(expected = ProtocolException.class)
	public void isOfferAcceptAgreementTest2() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, offer1).with(party1, accept2);
	}

	@Test
	public void testWithProgress() {
		AMOPState state = new AMOPState(Phase.INIT, actions, connections, null,
				settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		state = state.with(progresstime);
		assertEquals(progresstime, state.getProgress());
	}

	@Test(expected = IllegalArgumentException.class)
	public void testWithProgressAlreadySet() {
		AMOPState state = new AMOPState(Phase.INIT, actions, connections,
				progressrounds, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(progresstime);
	}

	@Test
	public void activePartiesWalkedAwayTest() {
		AMOPState state = new AMOPState(Phase.INIT, actions, connections3,
				progressrounds, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway3);
		assertEquals(Arrays.asList(party1, party2), state.getActiveParties());
	}

	@Test
	public void activePartiesAgreedTest() {
		Agreements agreements = new Agreements().with(new Agreements(bid1,
				new HashSet<>(Arrays.asList(party1, party2))));
		AMOPState state = new AMOPState(Phase.INIT, actions, connections3,
				progressrounds, settings, partyprofiles1, agreements,
				exceptions1, walkedaway0);
		assertEquals(Arrays.asList(party3), state.getActiveParties());
	}

	@Test
	public void activePartiesErrorTest() {
		Map<PartyId, ProtocolException> exceptions = Collections
				.singletonMap(party2, new ProtocolException("test", party2));
		AMOPState state = new AMOPState(Phase.INIT, actions, connections3,
				progressrounds, settings, partyprofiles1, agreements1,
				exceptions, walkedaway0);
		assertEquals(Arrays.asList(party1, party3), state.getActiveParties());
	}

	/************** isFinal tests *****************/
	@Test
	public void deadlineProgressTimeDeadlineTest() {
		Progress progress = mock(Progress.class);
		when(progress.isPastDeadline(any())).thenReturn(true);
		AMOPState state = new AMOPState(Phase.VOTE, actions, connections,
				progress, settings, partyprofiles1, agreements1, exceptions1,
				walkedaway0);
		assertTrue(state.isFinal(10));
	}

	@Test
	public void deadlineProgressTestOnly1Active() {
		Progress progress = mock(Progress.class);
		when(progress.isPastDeadline(any())).thenReturn(false);
		List<PartyId> walkedaway13 = Arrays.asList(party1, party3);

		AMOPState state = new AMOPState(Phase.VOTE, actions, connections,
				progress, settings, partyprofiles1, agreements1, exceptions1,
				walkedaway13);
		assertTrue(state.isFinal(10));
	}

	/*************
	 * action tests
	 ***************/
	@Test
	public void OfferTest() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, new Offer(party1, bid1));
		assertEquals(1, state.getActions().size());
	}

	@Test(expected = ProtocolException.class)
	public void OfferTwiceTest() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, new Offer(party1, bid1));
		// party1 may offer only once
		state = state.with(party1, new Offer(party1, bid1));
	}

	@Test
	public void endNegoOfferTest() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, new EndNegotiation(party1));
		assertEquals(1, state.getActions().size());
		assertEquals(1, state.getWalkedAway().size());
	}

	@Test
	public void endNegoVoteTest() throws ProtocolException {
		AMOPState state = new AMOPState(Phase.VOTE, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway0);
		state = state.with(party1, new EndNegotiation(party1));
		assertEquals(1, state.getActions().size());
		assertEquals(1, state.getWalkedAway().size());
	}

	@Test(expected = ProtocolException.class)
	public void OfferButInactive() throws ProtocolException {
		// party walked away or error'ed and then votes again
		AMOPState state = new AMOPState(Phase.OFFER, actions, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway1);
		state = state.with(party1, new Offer(party1, bid1));
	}

	/************** NextPhase tests *****************/
	@Test
	public void isInitToOfferWorking() throws ProtocolException {
		AMOPState state = state1.nextPhase();
		assertEquals(Phase.OFFER, state.getPhase());
	}

	@Test
	public void testCollectVotes() {
		Votes vote1AB = new Votes(party1, new HashSet<>(Arrays
				.asList(new Vote(party1, a, 2, 9), new Vote(party1, b, 2, 9))));
		Votes vote2AB = new Votes(party2, new HashSet<>(Arrays
				.asList(new Vote(party2, a, 2, 9), new Vote(party2, b, 2, 9))));
		Votes vote3C = new Votes(party3,
				Collections.singleton(new Vote(party3, c, 2, 9)));
		Votes vote4AC = new Votes(party4, new HashSet<>(Arrays
				.asList(new Vote(party4, a, 2, 9), new Vote(party4, c, 2, 9))));
		// party 1,2,4 vote for A, party 1,2 vote for B, party 3,4 vote for C.
		// the biggest vote is P,Q,S
		Map<PartyId, Votes> votes = new HashMap<>();
		votes.put(party1, vote1AB);
		votes.put(party2, vote2AB);
		votes.put(party3, vote3C);
		votes.put(party4, vote4AC);

		List<Action> actions1 = Arrays.asList(vote1AB, vote2AB, vote3C,
				vote4AC);
		AMOPState state = new AMOPState(Phase.VOTE, actions1, connections,
				progresstime, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway1);
		Agreements agrees = state.collectVotes(votes);
		System.out.println(agrees);

		// biggest agreement is party 1,2,4 for bid a.
		assertEquals(new HashSet<>(Arrays.asList(party1, party2, party4)),
				agrees.getMap().keySet());

	}

	@Test
	public void testNextPhase() {
		// copy of above. todo cleanup
		Votes vote1AB = new Votes(party1, new HashSet<>(Arrays
				.asList(new Vote(party1, a, 2, 9), new Vote(party1, b, 2, 9))));
		Votes vote2AB = new Votes(party2, new HashSet<>(Arrays
				.asList(new Vote(party2, a, 2, 9), new Vote(party2, b, 2, 9))));
		Votes vote3C = new Votes(party3,
				Collections.singleton(new Vote(party3, c, 2, 9)));
		Votes vote4AC = new Votes(party4, new HashSet<>(Arrays
				.asList(new Vote(party4, a, 2, 9), new Vote(party4, c, 2, 9))));
		// party 1,2,4 vote for A, party 1,2 vote for B, party 3,4 vote for C.
		// the biggest vote is P,Q,S
		Map<PartyId, Votes> votes = new HashMap<>();
		votes.put(party1, vote1AB);
		votes.put(party2, vote2AB);
		votes.put(party3, vote3C);
		votes.put(party4, vote4AC);

		List<Action> actions1 = Arrays.asList(vote1AB, vote2AB, vote3C,
				vote4AC);
		AMOPState state = new AMOPState(Phase.VOTE, actions1, connections,
				progressrounds, settings, partyprofiles1, agreements1,
				exceptions1, walkedaway1);
		state = state.nextPhase();
		assertEquals(Phase.OFFER, state.getPhase());
		assertEquals(new HashSet<>(Arrays.asList(party1, party2, party4)),
				state.getAgreements().getMap().keySet());
		assertEquals(progressrounds.advance(), state.getProgress());

	}
}
